"""
.. module:: Smoothing

Smoothing
*************

:Description: Smoothing

    

:Authors: bejar
    

:Version: 

:Created on: 21/07/2016 8:14 

"""

__author__ = 'bejar'

from scipy.sparse.linalg import spsolve
import numpy as np
import scipy as scp
from scipy import sparse
from scipy.sparse import linalg


def numpy_smoothing(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len - 1:0:-1], x, x[-1:-window_len:-1]]
    # print(len(s))
    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode='valid')
    return y[int(window_len / 2):int(-window_len / 2) + 1]


def ALS_smoothing(y, lam, p, niter=10):
    """
    Asymmetric Least Squares smoothing

    Method for smoothing also useful for baseline correction

    Taken from

    @article{eilers2005baseline,
     title={Baseline correction with asymmetric least squares smoothing},
     author={Eilers, Paul HC and Boelens, Hans FM},
     journal={Leiden University Medical Centre Report},
     year={2005}
    }

    :param y: signal
    :param lam: signal smoothing, usual values 10^2 - 10^9
    :param p: asymmetry usual values from 0.001 to 0.1 for baseline removal
                 (but for smoothing can be close to 0.9)
    :param niter: number of iterations,
    :return:
    """
    L = len(y)
    D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    w = np.ones(L)
    for i in range(niter):
        W = sparse.spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.transpose())
        z = spsolve(Z, w * y)
        w = p * (y > z) + (1 - p) * (y < z)
    return z


def tvdiplmax(y):
    """Calculate the value of lambda so that if lambda >= lambdamax, the TVD
    functional solved by TVDIP is minimized by the trivial constant solution
    x = mean(y). This can then be used to determine a useful range of values
    of lambda, for example.
    Args:
        y: Original signal to denoise, size N x 1.
    Returns:
        lambdamax: Value of lambda at which x = mean(y) is the output of the
            TVDIP function.
    """

    N = y.size
    M = N - 1

    # Construct sparse operator matrices
    I1 = sparse.eye(M)
    O1 = sparse.dia_matrix((M, 1))
    D = sparse.hstack([I1, O1]) - sparse.hstack([O1, I1])

    DDT = D.dot(D.conj().T)
    Dy = D.dot(y)

    lambdamax = np.absolute(linalg.spsolve(DDT, Dy)).max(0)

    return lambdamax


def tvdip(y, lambdas, display=1, stoptol=1e-3, maxiter=60):
    """Performs discrete total variation denoising (TVD) using a primal-dual
    interior-point solver. It minimizes the following discrete functional:
    E=(1/2)||y-x||_2^2+lambda*||Dx||_1
    over the variable x, given the input signal y, according to each value of
    the regularization parametero lambda > 0. D is the first difference matrix.
    Uses hot-restarts from each value of lambda to speed up convergence for
    subsequent values: best use of the feature is made by ensuring that the
    chosen lambda values are close to each other.
    Args:
        y: Original signal to denoise, size N x 1.
        lambdas: A vector of positive regularization parameters, size L x 1.
            TVD will be applied to each value in the vector.
        display: (Optional) Set to 0 to turn off progress display, 1 to turn
            on. Defaults to 1.
        stoptol: (Optional) Precision as determined by duality gap tolerance,
            if not specified defaults to 1e-3.
        maxiter: (Optional) Maximum interior-point iterations, if not specified
            defaults to 60.
    Returns:
        x: Denoised output signal for each value of lambda, size N x L.
        E: Objective functional at minimum for each lamvda, size L x 1.
        s: Optimization result, 1 = solved, 0 = maximum iterations
            exceeded before reaching duality gap tolerance, size L x 1.
        lambdamax: Maximum value of lambda for the given y. If
            lambda >= lambdamax, the output is the trivial constant solution
            x = mean(y).
    """

    # Search tuning parameters
    ALPHA = 0.01  # Backtracking linesearch parameter (0,0.5]
    BETA = 0.5  # Backtracking linesearch parameter (0,1)
    MAXLSITER = 20  # Max iterations of backtracking linesearch
    MU = 2  # t update

    N = y.size  # Length of input signal y
    M = N - 1  # Size of Dx

    # Construct sparse operator matrices
    I1 = sparse.eye(M)
    O1 = sparse.dia_matrix((M, 1))
    D = sparse.hstack([I1, O1]) - sparse.hstack([O1, I1])

    DDT = D.dot(D.conj().T)
    Dy = D.dot(y)

    # Find max value of lambda
    lambdamax = (np.absolute(linalg.spsolve(DDT, Dy))).max(0)

    if display:
        print("lambda_max=%5.2e" % lambdamax)

    L = lambdas.size
    x = np.zeros((N, L))
    s = np.zeros((L, 1))
    E = np.zeros((L, 1))

    # Optimization variables set up once at the start
    z = np.zeros((M, 1))
    mu1 = np.ones((M, 1))
    mu2 = np.ones((M, 1))

    # Work through each value of lambda, with hot-restart on optimization
    # variables
    for idx, l in enumerate(lambdas):
        t = 1e-10
        step = np.inf
        f1 = z - l
        f2 = -z - l

        # Main optimization loop
        s[idx] = 1

        if display:
            print("Solving for lambda={0:5.2e}, lambda/lambda_max={1:5.2e}".format(l, l / lambdamax))
            print("Iter# primal    Dual    Gap")

        for iters in range(maxiter):
            DTz = (z.conj().T * D).conj().T
            DDTz = D.dot(DTz)
            w = Dy - (mu1 - mu2)

            # Calculate objectives and primal-dual gap
            pobj1 = 0.5 * w.conj().T.dot(linalg.spsolve(DDT, w)) + l * (np.sum(mu1 + mu2))
            pobj2 = 0.5 * DTz.conj().T.dot(DTz) + l * np.sum(np.absolute(Dy - DDTz))
            pobj = np.minimum(pobj1, pobj2)
            dobj = -0.5 * DTz.conj().T.dot(DTz) + Dy.conj().T.dot(z)
            gap = pobj - dobj
            if display:
                print("{:5d} {:7.2e} {:7.2e} {:7.2e}".format(iters, pobj[0, 0],
                                                             dobj[0, 0],
                                                             gap[0, 0]))

            # Test duality gap stopping criterion
            if np.all(gap <= stoptol):  # ****
                s[idx] = 1
                break

            if step >= 0.2:
                t = np.maximum(2 * M * MU / gap, 1.2 * t)

            # Do Newton step
            rz = DDTz - w
            Sdata = (mu1 / f1 + mu2 / f2)
            S = DDT - sparse.csc_matrix((Sdata.reshape(Sdata.size),
                                         (np.arange(M), np.arange(M))))
            r = -DDTz + Dy + (1 / t) / f1 - (1 / t) / f2
            dz = linalg.spsolve(S, r).reshape(r.size, 1)
            dmu1 = -(mu1 + ((1 / t) + dz * mu1) / f1)
            dmu2 = -(mu2 + ((1 / t) - dz * mu2) / f2)

            resDual = rz.copy()
            resCent = np.vstack((-mu1 * f1 - 1 / t, -mu2 * f2 - 1 / t))
            residual = np.vstack((resDual, resCent))

            # Perform backtracking linesearch
            negIdx1 = dmu1 < 0
            negIdx2 = dmu2 < 0
            step = 1
            if np.any(negIdx1):
                step = np.minimum(step,
                                  0.99 * (-mu1[negIdx1] / dmu1[negIdx1]).min(0))
            if np.any(negIdx2):
                step = np.minimum(step,
                                  0.99 * (-mu2[negIdx2] / dmu2[negIdx2]).min(0))

            for _ in range(MAXLSITER):
                newz = z + step * dz
                newmu1 = mu1 + step * dmu1
                newmu2 = mu2 + step * dmu2
                newf1 = newz - l
                newf2 = -newz - l

                # Update residuals
                newResDual = DDT.dot(newz) - Dy + newmu1 - newmu2
                newResCent = np.vstack((-newmu1 * newf1 - 1 / t, -newmu2 * newf2 - 1 / t))
                newResidual = np.vstack((newResDual, newResCent))

                if (np.maximum(newf1.max(0), newf2.max(0)) < 0
                        and (scp.linalg.norm(newResidual) <=
                             (1 - ALPHA * step) * scp.linalg.norm(residual))):
                    break

                step = BETA * step

            # Update primal and dual optimization parameters
            z = newz
            mu1 = newmu1
            mu2 = newmu2
            f1 = newf1
            f2 = newf2

        x[:, idx] = (y - D.conj().T.dot(z)).reshape(x.shape[0])
        xval = x[:, idx].reshape(x.shape[0], 1)
        E[idx] = 0.5 * np.sum((y - xval) ** 2) + l * np.sum(np.absolute(D.dot(xval)))

        # We may have a close solution that does not satisfy the duality gap
        if iters >= maxiter:
            s[idx] = 0

        if display:
            if s[idx]:
                print("Solved to precision of duality gap %5.2e") % gap
            else:
                print("Max iterations exceeded - solution may be inaccurate")

    return x, E, s, lambdamax



"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

    

:Authors: bejar
    

:Version: 

:Created on: 30/06/2017 12:35 

"""

__author__ = 'bejar'

from .MFT import mft
from .STFT import stft

__all__ = ['mft', 'stft']


"""
.. module:: SFFT

SFFT
*************

:Description: SFFT

    

:Authors: bejar
    

:Version: 

:Created on: 19/06/2015 11:36 

"""

__author__ = 'bejar'

import scipy
import numpy as np


def stft(x, fftsize=1024, overlap=4, ban=0):
    """
    Short Time Fourier Transform

    :param x: Signal
    :param fftsize: Window length
    :param overlap: Overlaping between consecutive frequencies
    :param ban: numer of Frequencies to null
    :return:
    """
    hop = int(fftsize / overlap)
    w = scipy.hanning(fftsize + 1)[:-1]  # better reconstruction with this trick +1)[:-1]
    l = []
    for i in range(0, len(x) - fftsize, hop):
        v = np.fft.rfft(w * x[i:i + fftsize])
        for j in range(ban):
            v[j] = 0
        l.append(np.abs(v) ** 2 / np.max(np.abs(v) ** 2))
    return np.array(l)



"""
.. module:: SimpleConsensusClustering

SimpleConsensusClustering
*************

:Description: SimpleConsensusClustering

    

:Authors: bejar
    

:Version: 

:Created on: 22/01/2015 10:46 

"""

__author__ = 'bejar'


import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.cluster import KMeans, SpectralClustering
from numpy.random import randint

class SimpleConsensusClustering(BaseEstimator, ClusterMixin, TransformerMixin):
    """Simple Consensus Clustering Algorithm

    Pararemeters:

    n_clusters: int
        Number of clusters of the base clusterers and the consensus cluster
    base: string
        base clusterer ['kmeans']
    n_components: int
        number of components of the consensus
    consensus: string
        consensus method ['coincidence']

    """
    def __init__(self, n_clusters, n_clusters_base=None, ncb_rand=False, base='kmeans', n_components=10,
                 consensus='coincidence', consensus2='kmeans'):
        self.n_clusters = n_clusters
        if n_clusters_base is None:
            self.n_clusters_base = n_clusters
        else:
            self.n_clusters_base = n_clusters_base
        self.ncb_rand = ncb_rand
        self.cluster_centers_ = None
        self.labels_ = None
        self.cluster_sizes_ = None
        self.base = base
        self.n_components = n_components
        self.consensus = consensus
        self.consensus2 = consensus2


    def fit(self, X):
        """
        Clusters the examples
        :param X:
        :return:
        """

        if self.consensus == 'coincidence':
            self.cluster_centers_, self.labels_ = self._fit_process_coincidence(X)

    def _fit_process_coincidence(self, X):
        """
        Obtains n_components kmeans clustering, compute the coincidence matrix and applies kmeans to that coincidence
        matrix

        :param X:
        :return:
        """
        baseclust = []

        for i in range(self.n_components):
            ncl = self.n_clusters_base if self.ncb_rand else randint(2, self.n_clusters_base+1)
            if self.base == 'kmeans':
                km = KMeans(n_clusters=ncl, n_init=1, init='random')
            elif self.base == 'spectral':
                km = SpectralClustering(n_clusters=ncl, assign_labels='discretize',
                                        affinity='nearest_neighbors', n_neighbors=30)
            km.fit(X)
            baseclust.append(km.labels_)

        coin_matrix = np.zeros((X.shape[0], X.shape[0]))

        for l in baseclust:
            for i in range(X.shape[0]):
                coin_matrix[i, i] += 1
                for j in range(i+1, X.shape[0]):
                    #if i != j:
                        if l[i] == l[j]:
                            coin_matrix[i, j] += 1
                            coin_matrix[j, i] += 1

        coin_matrix /= self.n_components
        if self.consensus2 == 'kmeans':
            kmc = KMeans(n_clusters=self.n_clusters, n_jobs=-1)
            kmc.fit(coin_matrix)
            return kmc.cluster_centers_, kmc.labels_
        elif self.consensus2 == 'spectral':
            kmc = SpectralClustering(n_clusters=self.n_clusters, assign_labels='discretize',
                                    affinity='nearest_neighbors', n_neighbors=40)
            kmc.fit(coin_matrix)

            return None, kmc.labels_


if __name__ == '__main__':
    from sklearn import datasets
    from sklearn.metrics import adjusted_mutual_info_score
    from kemlglearn.datasets import make_blobs
    import matplotlib.pyplot as plt


    # data = datasets.load_iris()['data']
    # labels = datasets.load_iris()['target']

    # data, labels = make_blobs(n_samples=[100, 200], n_features=2, centers=[[1,1], [0,0]], random_state=2, cluster_std=[0.2, 0.4])
    data, labels = datasets.make_circles(n_samples=400, noise=0.1, random_state=4, factor=0.3)

    km = KMeans(n_clusters=2)

    cons = SimpleConsensusClustering(n_clusters=2, n_clusters_base=20, n_components=50, ncb_rand=False)

    lkm = km.fit_predict(data)
    cons.fit(data)
    lcons = cons.labels_

    print(adjusted_mutual_info_score(lkm, labels))
    print(adjusted_mutual_info_score(lcons, labels))

    fig = plt.figure()

    # ax = fig.gca(projection='3d')
    # pl.scatter(X[:, 1], X[:, 2], zs=X[:, 0], c=ld.labels_, s=25)
    #
    ax = fig.add_subplot(131)
    plt.scatter(data[:,0],data[:,1],c=labels)
    ax = fig.add_subplot(132)
    plt.scatter(data[:,0],data[:,1],c=lkm)
    ax = fig.add_subplot(133)
    plt.scatter(data[:,0],data[:,1],c=lcons)

    plt.show()



"""
.. module:: KMedoidsFlexible

KMedoidsFlexible
*************

:Description: KMedoidsFlexible

    

:Authors: bejar
    

:Version: 

:Created on: 18/12/2017 10:45 

"""

from __future__ import division

__author__ = 'bejar'

from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import euclidean_distances
from numpy.random import choice
import numpy as np


class KMedoidsFlexible(BaseEstimator, ClusterMixin, TransformerMixin):
    """K-Medoids algorithm with distance as a parameter

    Partitions a dataset using the K-medoids algorithm with arbitrary distance (passed as a parameter)

    Parameters:

    n_clusters: int
    max_iter: int
    tol: float
    distance: function or "precomputed"

       receives a distance function for the computation able to return array x array distances, it Precomputes the
       distance matrix for faster computation when fit/fit_predict is called, but needs a quadratic amount of memory
       so can not be used for large datasets

       if it is "precomputed" then the data is assumed to be the distance matrix (in condensed form, upper triangular)
    """

    def __init__(self, n_clusters=3, max_iter=50, distance=euclidean_distances, random_state=None):
        self.cluster_medoids_ = None
        self.labels_ = None
        self.distance_ = distance
        self.nclusters_ = n_clusters
        self.max_iter_ = max_iter
        self.distance_matrix_ = None
        self.random_ = random_state

    def fit(self, X):
        """
        Clusters the examples
        :param X:
        :return:
        """
        if self.distance_ == 'precomputed':
            raise NameError('For precomputed distance use fit_predict')
        else:
            self._fit_process(X)
            return self

    def fit_predict(self, X):
        """
        Clusters the examples
        :param X:
        :return:
        """
        self.labels_ = self._fit_process(X)

        return self.labels_


    def predict(self, X):
        """
        Returns the nearest cluster for a data matrix

        @param X:
        @return:
        """
        if self.distance_ == 'precomputed':
            raise NameError('Cannot predict for new data with precomputed distance')
        else:
            clasif = []
            for i in range(X.shape[0]):
                clasif.append(np.argmin(self.distance_(X[i].reshape(1, -1), self.cluster_medoids_)))
            return clasif

    def _fit_process(self, X):
        """
        Computes K-medoids
        :param X:
        :return:
        """

        if self.distance_ == 'precomputed':
            self.distance_matrix_ = X
            self.nexamp_ = int(np.sqrt(X.shape[0] * 2)) + 1
        else:  # A distance function
            self.nexamp_ = X.shape[0]
            self.distance_matrix_ = np.zeros(self.nexamp_ * (self.nexamp_ - 1) // 2)
            # Condensed distance matrix computation
            for i in range(self.nexamp_):
                for j in range(i+1, self.nexamp_):
                    self.distance_matrix_[self.sel(i, j)] = self.distance_(X[i].reshape(1, -1), X[j].reshape(1, -1))

        # TODO: Implement K-means++ initialization strategy
        # Initializes the medoids randomly
        medoids = np.array(choice(range(self.nexamp_), self.nclusters_, replace=False), dtype=int)
        assignments = np.zeros(self.nexamp_, dtype=int)
        for j in range(self.nexamp_):
            assignments[j] = self._find_nearest_medoid(j, medoids)

        for i in range(self.max_iter_):
            new_medoids = self._kmedoids_iter(assignments)

            if np.array_equal(new_medoids, medoids):
                break
            else:
                medoids = new_medoids
            for j in range(self.nexamp_):
                assignments[j] = self._find_nearest_medoid(j, new_medoids)

        if self.distance_ == 'precomputed':
            # If distance is precomputed we store the indices corresponding to the medoids
            self.cluster_medoids_ = medoids
        else:
            self.cluster_medoids_ = X[medoids, :]

        return assignments

    def _kmedoids_iter(self, assignments):
        """
        One iteration of k medoids
        :param X:
        :param medoids:
        :return:
        """
        lassign = [[] for i in range(self.nclusters_)]
        for i in range(self.nexamp_):
            lassign[assignments[i]].append(i)

        # Compute the new medoids
        new_medoids = np.zeros(self.nclusters_, dtype=int)
        for i, ass in enumerate(lassign):
            mndist = np.inf
            for a1 in ass:
                mddist = np.sum([self.sel_distance(a1, c) for c in ass])
                if mddist < mndist:
                    new_medoids[i] = a1
                    mndist = mddist

        return new_medoids

    def _find_nearest_medoid(self, examp, centers):
        """
        Finds the nearest cluster for an example
        :param examp:
        :param centers:
        :return:
        """
        return np.argmin([self.sel_distance(examp, c) for c in centers])

    def sel(self, i, j):
        return self.nexamp_ * i - (i * (i+1)//2) + (j - i) - 1

    def sel_distance(self, i, j):
        """
        Selects the distance between two examples from the distance matrix

        :param i: int
        :param j: int
        :return: float
        """
        if i == j:
            return 0
        elif i < j:
            return self.distance_matrix_[self.nexamp_ * i - (i * (i+1)//2) + (j -i) - 1]
        else:
            return self.distance_matrix_[self.nexamp_ * j - (j * (j+1)//2) + (i -j) - 1]


if __name__ == '__main__':
    from kemlglearn.datasets import make_blobs
    from scipy.spatial import distance
    from sklearn.metrics import pairwise
    import matplotlib.pyplot as plt
    from numpy.random import normal

    sc1=75
    v1=0.1
    sc2=75
    v2=0.9
    X = np.zeros((sc1 + sc2, 2))
    X[0:sc1, 0] = normal(loc=-0.5, scale=v1, size=sc1)
    X[0:sc1, 1] = normal(loc=0.0, scale=v2, size=sc1)
    X[sc1:, 0] = normal(loc=0.5, scale=v1, size=sc2)
    X[sc1:, 1] = normal(loc=0.0, scale=v2, size=sc2)
    dlabels = np.zeros(sc1 + sc2)
    dlabels[sc1:] = 1
    cosine = lambda x,y: distance.cdist(x,y, metric='braycurtis')


    # km = KMedoidsFlexible(n_clusters=2, distance=cosine)
    # labels = km.fit_predict(X)
    km = KMedoidsFlexible(n_clusters=2, distance='precomputed')
    D = cosine(X,X)
    D = distance.squareform(D, force='tovector', checks=False)
    labels = km.fit_predict(D)


    fig = plt.figure()

    ax = fig.add_subplot(111)
    plt.scatter(X[:, 0], X[:, 1], c=labels)
    medoids = km.cluster_medoids_

    for i, m in enumerate(medoids):
        # plt.scatter(medoids[i, 0], medoids[i, 1], c=i, marker='x', s=200)
        plt.scatter(X[m, 0], X[m, 1], c=i, marker='x', s=200)

    plt.ylim(-3,3)
    plt.xlim(-3,3)
    plt.show()

    print(labels)


"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

    

:Authors: bejar
    

:Version: 

:Created on: 18/03/2015 13:49 

"""

__author__ = 'bejar'


from .SAX import SAX
from .BOSS import Boss

__all__ = ['SAX', 'Boss']


"""
.. module:: BOSS

BOSS
*************

:Description: BOSS

    

:Authors: bejar
    

:Version: 

:Created on: 15/02/2017 13:48 

"""


import numpy as np
from kemlglearn.preprocessing import Discretizer
import seaborn as sn
from collections import Counter
from kemlglearn.time_series.decomposition.MFT import mft
__author__ = 'bejar'


def boss_distance(histo1, histo2):
    """
    BOSS distance between two histograms

    Not really a distance because it is not symmetric

    :param histo1:
    :param histo2:
    :return:
    """
    val = 0
    for w in histo1:
        if w in histo2:
            val += (histo1[w] - histo2[w]) ** 2
        else:
            val += histo1[w] * histo1[w]

    return val


def euclidean_distance(histo1, histo2):
    """
    Euclidean distance between two histograms

    :param histo1:
    :param histo2:
    :return:
    """
    val = 0
    lnot = []
    for w in histo1:
        if w in histo2:
            val += (histo1[w] - histo2[w]) ** 2
            lnot.append(w)
        else:
            val += histo1[w] * histo1[w]

    for w in histo2:
        if w not in lnot:
            val += histo2[w] * histo2[w]

    return val


def cosine_similarity(histo1, histo2):
    """
    Euclidean distance between two histograms

    :param histo1:
    :param histo2:
    :return:
    """
    val = 0.0
    norm1 = 0.0
    for w in histo1:
        if w in histo2:
            val += (histo1[w] * histo2[w])
        norm1 += histo1[w] ** 2
    norm2 = 0.0
    for w in histo2:
        norm2 += histo2[w] ** 2

    return val / (np.sqrt(norm1) * np.sqrt(norm2))


def hamming_distance(histo1, histo2):
    """
    Euclidean distance between two histograms

    :param histo1:
    :param histo2:
    :return:
    """
    val = 0
    lnot = []
    for w in histo1:
        if w in histo2:
            val += np.abs((histo1[w] - histo2[w]))
            lnot.append(w)
        else:
            val += histo1[w]

    for w in histo2:
        if w not in lnot:
            val += histo2[w]

    return val


def bin_hamming_distance(histo1, histo2):
    """
    jaccard distance between two histograms

    :param histo1:
    :param histo2:
    :return:
    """
    s1 = set(histo1.keys())
    s2 = set(histo2.keys())
    return len(s1) + len(s2) - len(s1.intersection(s2))


class Boss():
    """
    Computes the BOSS words for a dictionary of series
    """

    def __init__(self, dseries, sampling, butfirst=False):
        """

        :param dseries:
        :param sampling:
        """
        self.series = dseries
        self.sampling = sampling
        self.coefs = {}
        self.codes = {}
        self.butfirst = butfirst

    def discretization_intervals(self, ncoef, wsize, vsize):
        """
        Computes the BOSS discretization for the signals, the word length is 2*ncoefs (real and imaginary part) except
        is there are coefficients that are zero

        :param ncoef:
        :param wsize:
        :return:
        """

        all_coefs = []
        for s in self.series:
            coefs = mft(self.series[s], self.sampling, ncoef, wsize, butfirst=self.butfirst)
            lcoefs = []
            for i in range(coefs.shape[1]):
                lcoefs.append(coefs[:, i].real)
                lcoefs.append(coefs[:, i].imag)

            all_coefs.append(np.stack(lcoefs, axis=-1))
            self.coefs[s] = all_coefs[-1]

        X = np.concatenate(all_coefs)

        self.disc = Discretizer(method='frequency', bins=vsize)
        self.disc.fit(X)

    def discretize(self):
        """
        Computes the words for each time series
        :param series:
        :return:
        """
        vocabulary = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

        def word(vec):
            """

            :param v:
            :return:
            """
            w = ''
            for v in vec:
                w += vocabulary[int(v)]
            return w

        for c in self.coefs:
            sdisc = self.disc.transform(self.coefs[c], copy=True).real
            prevw = word(sdisc[0])
            lvoc = [prevw]
            for i in range(1, sdisc.shape[0]):
                nword = word(sdisc[i])
                if nword != prevw:
                    lvoc.append(nword)
            self.codes[c] = Counter(lvoc)

            # print(c, self.codes[c])


if __name__ == '__main__':
    pass
    # wlen = 64
    # voclen = 3
    # ncoefs = 3
    #

    #
    # boss = Boss(dseries, 10)
    # boss.discretization_intervals(ncoefs, wlen, voclen)
    # boss.discretize()



"""
.. module:: test

test
*************

:Description: test

    

:Authors: bejar
    

:Version: 

:Created on: 07/07/2014 11:12 

"""
__author__ = 'bejar'



# from Leader import Leader
# from GlobalKMeans import GlobalKMeans
from sklearn.datasets import load_iris, make_circles
import pylab as pl
from kemlglearn.datasets import make_blobs
import matplotlib.pyplot as plt
# from sklearn.cluster import KMeans
# from kemlglearn.datasets import cluster_generator
# from kemlglearn.metrics import within_scatter_matrix_score, between_scatter_matrix_score, calinski_harabasz_score,\
#     zhao_chu_franti_score, scatter_matrices_scores, davies_bouldin_score, variation_of_information_score, \
#     jaccard_score, JaccardArandFolkes, folkes_mallow_score
# from kemlglearn.cluster.consensus import SimpleConsensusClustering
# from sklearn.metrics.cluster import normalized_mutual_info_score

X, y_data = make_blobs(n_samples=[25, 200], n_features=2, centers=[[1,1], [0,0]], random_state=2, cluster_std=[0.1, 0.4])
# X , y_data= load_iris(return_X_y=True)

# X, y_data = make_circles(n_samples=1000, noise=0.5, random_state=4, factor=0.5)


# nc = 12
# _, X = cluster_generator(n_clusters=nc, sepval=0.01, numNonNoisy=15, numNoisy=3, rangeN=[50, 100])
# ld = Leader(radius=25.0)
#
# #print timeit.timeit(stmt='ld.fit(X)',setup=setup,number=10)
# ld.fit(X)
#
# # print ld.cluster_centers_.shape[0]
#
# fig = plt.figure()
#
# # ax = fig.gca(projection='3d')
# # pl.scatter(X[:, 1], X[:, 2], zs=X[:, 0], c=ld.labels_, s=25)
# #
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=ld.labels_)
#
# plt.show()

# gkm = GlobalKMeans(n_clusters=nc, algorithm='bagirov')
# gkm.fit(X)
# print DaviesBouldin(X, gkm.labels_)
# print scatter_matrices_scores(X, gkm.labels_, ['Inertia'])

#
# print gkm.cluster_centers_.shape[0], gkm.inertia_
# print within_scatter_matrix_score(X, gkm.labels_)
# print between_scatter_matrix_score(X, gkm.labels_)


# fig = plt.figure()
#
# # ax = fig.gca(projection='3d')
# # pl.scatter(X[:, 1], X[:, 2], zs=X[:, 0], c=gkm.labels_, s=25)
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:, 0], X[:, 1], c=gkm.labels_)
#
# plt.show()
# for nc in range(2, 16):
#     km = KMeans(n_clusters=nc)
#     km.fit(X)
#     #print km.cluster_centers_.shape[0], km.inertia_
#     print nc, scatter_matrices_scores(X, km.labels_, ['CH', 'ZCF', 'Hartigan', 'Xu'])
#     print DaviesBouldin(X, km.labels_)

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=km.labels_)
#
# plt.show()

# km = KMeans(n_clusters=3)
# km.fit(X)
#
# # print variation_of_information_score(km.labels_, y_data)
# # print normalized_mutual_info_score(km.labels_, y_data)
# print(jaccard_score(km.labels_, y_data))
# print(folkes_mallow_score(km.labels_, y_data))
# print(JaccardArandFolkes(km.labels_, y_data))

# print DaviesBouldin(X, km.labels_)
# print scatter_matrices_scores(X, km.labels_, ['Inertia'])

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=km.labels_)
#
# plt.show()


# simple = SimpleConsensusClustering(n_clusters=nc, n_components=40)
# simple.fit(X)
#
# print DaviesBouldin(X, simple.labels_)
# print scatter_matrices_scores(X, simple.labels_, ['Inertia'])

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=simple.labels_)
#
# plt.show()


# simple = SimpleConsensusClustering(n_clusters=nc, n_components=40, consensus2='spectral')
# simple.fit(X)
#
# print DaviesBouldin(X, simple.labels_)
# print scatter_matrices_scores(X, simple.labels_, ['Inertia'])
#

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:, 0], X[:, 1], c=y_data)
#
# plt.show()
from numpy.random import normal
import numpy as np
sc1=75
v1=0.1
sc2=75
v2=0.9
X = np.zeros((sc1 + sc2, 2))
X[0:sc1, 0] = normal(loc=0.0, scale=v1, size=sc1)
X[0:sc1, 1] = normal(loc=0.0, scale=v2, size=sc1)
X[sc1:, 0] = normal(loc=1, scale=v1, size=sc2)
X[sc1:, 1] = normal(loc=0.0, scale=v2, size=sc2)
dlabels = np.zeros(sc1 + sc2)
dlabels[sc1:] = 1

from .KMedoidsFlexible import KMedoidsFlexible

km = KMedoidsFlexible(n_clusters=2)

labels = km.fit_predict(X)

fig = plt.figure()

ax = fig.add_subplot(111)
plt.scatter(X[:, 0], X[:, 1], c=labels)
medoids = km.cluster_medoids_

for i, m in enumerate(medoids):
    plt.scatter(medoids[i, 0], medoids[i, 1], c=i, marker='x', s=200)

plt.show()

labels = km.predict(X)

print(labels)



"""
.. module:: MFT

MFT
*************

:Description: MFT

    Computes the Momentary Fourier Transformation

    Computes the n firsts coefficients of the FFT for consecutive windows given the n first FFT coeffs of the first window
    (cost is O(nwindows*ncoef) instead of  O(wsize*log(wsize)*nwindows)

    Albrecht, Cumming, Dudas "The Momentary Fourier Transformation Derived from Recursive Matrix Transformation"

@inproceedings{albrecht1997momentary,
  title={The momentary fourier transformation derived from recursive matrix transformations},
  author={Albrecht, S and Cumming, I and Dudas, J},
  booktitle={Digital Signal Processing Proceedings, 1997. DSP 97., 1997 13th International Conference on},
  volume={1},
  pages={337--340},
  year={1997},
  organization={IEEE}
}

:Authors: bejar
    

:Version: 

:Created on: 15/02/2017 13:48 

"""
import numpy as np
import time

__author__ = 'bejar'


def mft(series, sampling, ncoef, wsize, butfirst=False):
    """
    Computes the ncoef fourier coefficient for the series

    butfirst eliminates the first coefficient (mean of the signal)

    :return:
    """

    if butfirst:
        ncoef += 1
    nwindows = len(series) - wsize
    # imaginary matrix for the coefficients
    coef = np.zeros((nwindows, ncoef), dtype=np.complex)
    dcoef = np.zeros(ncoef, dtype=np.complex)
    fwsize = float(wsize)
    pi2i = 1j * 2 * np.pi

    y = np.fft.rfft(series[:wsize])
    for l in range(ncoef):
        coef[0, l] = y[l]
        dcoef[l] = np.exp(pi2i * (l / fwsize))

    for w in range(1, nwindows):
        coef[w] = dcoef * (coef[w - 1] + (series[wsize + (w - 1)] - series[w - 1]))

    if butfirst:
        return coef[:, 1:]
    else:
        return coef



if __name__ == '__main__':
    pass
    # itime = time.time()
    # coef1 = mft(signal, sampling=10, ncoef=15, wsize=32)
    # ftime = time.time()
    # print(ftime - itime)




from distutils.core import setup

setup(
    name='kemlglearn',
    version='0.4.1',
    packages=['kemlglearn',
              'kemlglearn.cluster',
              'kemlglearn.cluster.consensus',
              'kemlglearn.cluster.border',
              'kemlglearn.metrics',
              'kemlglearn.datasets',
              'kemlglearn.feature_selection',
              'kemlglearn.feature_selection.unsupervised',
              'kemlglearn.preprocessing',
              'kemlglearn.time_series',
              'kemlglearn.time_series.discretization',
              'kemlglearn.time_series.decomposition',
              'kemlglearn.time_series.smoothing'],
    url='http://www.cs.upc.edu/~bejar',
    license='GPL 2.0',
    author='Javier Bejar',
    author_email='bejar@cs.upc.edu',
    description='Machine Learning algorithms/functions '
)



"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

    

:Authors: bejar
    

:Version: 

:Created on: 16/06/2016 9:23 

"""

from .BorderDetection import QiuCaoBorder

__author__ = 'bejar'

__all__ = ['QiuCaoBorder']